@@ -706,10 +706,11 @@ cfg = ConfigParser.RawConfigParser()
 cfg.optionxform = str
 
 valid_ranges = {
-	"3dlut.hdr_peak_luminance": [120.0, 10000.0],
+	"3dlut.hdr_peak_luminance": [100.0, 10000.0],
 	"3dlut.hdr_minmll": [0.0, 0.1],
-	"3dlut.hdr_maxmll": [120.0, 10000.0],
+	"3dlut.hdr_maxmll": [100.0, 10000.0],
 	"3dlut.trc_gamma": [0.000001, 10],
+	"3dlut.hdr_sat": [0.0, 1.0],
 	"3dlut.trc_output_offset": [0.0, 1.0],
 	"app.port": [1, 65535],
 	"gamma": [0.000001, 10],
@@ -847,8 +848,10 @@ defaults = {
 	"3dlut.hdr_display": 0,
 	"3dlut.hdr_minmll": 0.0,
 	"3dlut.hdr_maxmll": 10000.0,
-	"3dlut.hdr_peak_luminance": 400.0,
+	"3dlut.hdr_maxmll_alt_clip": 1,
+	"3dlut.hdr_peak_luminance": 480.0,
 	"3dlut.hdr_ambient_luminance": 5.0,
+	"3dlut.hdr_sat": 0.5,
 	"3dlut.image.layout": "h",
 	"3dlut.image.order": "rgb",
 	"3dlut.input.profile": "",
@@ -1634,7 +1637,7 @@ def initcfg(module=None):
 		cfgnames.append(cfgbasename)
 	else:
 		cfgnames.extend("%s-%s" % (appbasename, othermod) for othermod in
-						("synthprofile", "testchart-editor"))
+						("testchart-editor", ))
 	cfgroots = [confighome]
 	if module == "apply-profiles":
 		cfgroots.append(config_sys)
@@ -1659,7 +1662,7 @@ def initcfg(module=None):
 		safe_print("Warning - could not parse configuration files:\n%s" %
 				   "\n".join(cfgfiles))
 		# Fix Python 2.7 ConfigParser option values being lists instead of
-		# strings in case of a ParsingError. http://bugs.python.org/issue2414
+		# strings in case of a ParsingError. http://bugs.python.org/issue24142
 		all_sections = [ConfigParser.DEFAULTSECT]
 		all_sections.extend(cfg.sections())
 		for section in all_sections:
