@@ -23,9 +23,12 @@ if sys.platform == "win32":
 if sys.platform == "win32":
 	import win32api
 
+import ICCProfile as ICCP
 import localization as lang
+from imfile import tiff_get_header
 from log import safe_print as log_safe_print
 from meta import name as appname, version
+from network import get_network_addr, get_valid_host
 from ordereddict import OrderedDict
 from util_str import safe_str, safe_unicode
 
@@ -35,7 +38,7 @@ CALLBACK = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(None),
 							ctypes.c_char_p, ctypes.c_ulonglong, ctypes.c_bool)
 
 
-min_version = (0, 87, 14, 0)
+min_version = (0, 88, 14, 0)
 
 
 # Search for madTPG on the local PC, connect to the first found instance
@@ -52,20 +55,20 @@ CM_ShowIpAddrDialog = 4
 CM_Fail = 5
 
 
-_methodnames = ("ConnectEx", "Disable3dlut", "Enable3dlut",
+_methodnames = ("ConnectEx", "Disable3dlut", "Enable3dlut", "EnterFullscreen",
 				"GetBlackAndWhiteLevel", "GetDeviceGammaRamp",
 				"GetSelected3dlut", "GetVersion",
-				"IsDisableOsdButtonPressed",
+				"IsDisableOsdButtonPressed", "IsFullscreen",
 				"IsStayOnTopButtonPressed",
-				"IsUseFullscreenButtonPressed",
+				"IsUseFullscreenButtonPressed", "LeaveFullscreen",
 				"SetDisableOsdButton",
 				"SetDeviceGammaRamp", "SetOsdText",
 				"GetPatternConfig", "SetPatternConfig",
 				"ShowProgressBar", "SetProgressBarPos",
 				"SetSelected3dlut", "SetStayOnTopButton",
 				"SetUseFullscreenButton", "ShowRGB",
-				"ShowRGBEx", "Load3dlutFile", "Disconnect",
-				"Quit", "Load3dlutFromArray256")
+				"ShowRGBEx", "Load3dlutFile", "LoadHdr3dlutFile", "Disconnect",
+				"Quit", "Load3dlutFromArray256", "LoadHdr3dlutFromArray256")
 
 _autonet_methodnames = ("AddConnectionCallback", "Listen", "Announce")
 
@@ -109,10 +112,16 @@ class H3DLUT(object):
 
 	# https://sourceforge.net/projects/thr3dlut
 
-	def __init__(self, filename):
-		self.fileName = filename
-		with open(filename, "rb") as lut:
-			data = lut.read()
+	def __init__(self, stream_or_filename=None, check_lut_size=True):
+		if not stream_or_filename:
+			return
+		if isinstance(stream_or_filename, basestring):
+			self.fileName = stream_or_filename
+			with open(stream_or_filename, "rb") as lut:
+				data = lut.read()
+		else:
+			self.fileName = None
+			data = stream_or_filename.read()
 		self.signature = data[:4]
 		self.fileVersion = struct.unpack("<l", data[4:8])[0]
 		self.programName = data[8:40].rstrip("\0")
@@ -135,7 +144,7 @@ class H3DLUT(object):
 								   parametersSize]
 		self.LUTDATA = data[self.lutFileOffset:
 							self.lutFileOffset + self.lutCompressedSize]
-		if len(self.LUTDATA) != self.lutCompressedSize:
+		if check_lut_size and len(self.LUTDATA) != self.lutCompressedSize:
 			raise ValueError("3DLUT size %i does not match expected size %i" %
 							 (len(self.LUTDATA), self.lutCompressedSize))
 		if len(data) == self.lutFileOffset + self.lutCompressedSize + 1552:
@@ -169,32 +178,167 @@ class H3DLUT(object):
 	def parametersSize(self):
 		return len(self.parametersData)
 
+	def _get_stream(self, stream_or_filename=None, ext=None):
+		if not stream_or_filename:
+			stream_or_filename = self.fileName
+			if ext:
+				stream_or_filename = os.path.splitext(stream_or_filename)[0] + ext
+		if isinstance(stream_or_filename, basestring):
+			stream = open(stream_or_filename, "wb")
+		else:
+			stream = stream_or_filename
+		return stream
+
 	def write(self, stream_or_filename=None):
 		"""
 		Write 3D LUT to stream or filename.
 		
 		"""
-		if not stream_or_filename:
-			stream_or_filename = self.fileName
+		stream = self._get_stream(stream_or_filename)
+		stream.write(self.data)
 		if isinstance(stream_or_filename, basestring):
-			stream = open(stream_or_filename, "wb")
 			if not self.fileName:
 				self.fileName = stream_or_filename
+			stream.close()
+
+	def write_devicelink(self, stream_or_filename=None):
+		"""
+		Write 3D LUT to ICC device link.
+		
+		"""
+		stream = self._get_stream(stream_or_filename, ".icc")
+
+		link = ICCP.ICCProfile()
+		link.connectionColorSpace = "RGB"
+		link.profileClass = "link"
+		link.tags.desc = ICCP.TextDescriptionType()
+		link.tags.desc.ASCII = os.path.splitext(os.path.basename(stream.name))[0]
+		link.tags.cprt = ICCP.TextType("text\0\0\0\0No copyright", "cprt")
+
+		input_grid_steps = 2 ** self.inputBitDepth[0]  # Assume equal bitdepth for R, G, B
+		if input_grid_steps > 255:
+			# madVR 3D LUTs are 256^3, but ICC LUT16Type only supports up to
+			# 255^3. As madVR 3D LUTs use video levels encoding, we simply skip
+			# the first cLUT entry in each dimension and fix the offset by
+			# scaling the input/output shaper curves. That way, only level 1 of
+			# 255 will be affected (with black at 16 and white at 235),
+			# which isn't used in actual video content.
+			clut_grid_steps = 255
 		else:
-			stream = stream_or_filename
-		stream.write(self.data)
+			clut_grid_steps = input_grid_steps
+		# Filling a 255^3 list is VERY memory intensive in Python, so we 'fake'
+		# the LUT16Type cLUT and only use tag data of offsets/sizes and shaper
+		# curves while writing the raw cLUT data directly without going through
+		# decoding/re-encoding roundtrip
+		A2B0 = ICCP.LUT16Type()
+		A2B0.matrix = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
+		A2B0.input = []
+		for i in xrange(3):
+			A2B0.input.append([])
+			for j in xrange(4096):
+				A2B0.input[-1].append(min(max(j / 4095. * (256 / 255.) - (256 / 255. - 1), 0), 1) * 65535)
+		input_bytes = len(A2B0.input) * len(A2B0.input[0]) * 2
+		A2B0.clut = [[[0] * 3 for i in xrange(clut_grid_steps)]]  # Fake cLUT
+		A2B0.output = []
+		for i in xrange(3):
+			A2B0.output.append([])
+			for j in xrange(4096):
+				A2B0.output[-1].append(min(max(j / 4095. * (256 / 255.), 0), 1) * 65535)
+		output_bytes = len(A2B0.output) * len(A2B0.output[0]) * 2
+		tagData = A2B0.tagData[:52 + input_bytes]  # Exclude cLUT and output curves
+
+		# Write actual cLUT
+		# XXX Currently only 16 bit RGB data is supported
+		samples_per_pixel = 3  # RGB
+		bytes_per_sample = self.outputBitDepth / 8
+		bytes_per_pixel = samples_per_pixel * bytes_per_sample
+		io = StringIO(tagData)
+		io.seek(0, 2)  # Position cursor at end
+		i = 0
+		for R in xrange(input_grid_steps):
+			if not R:
+				i += input_grid_steps * input_grid_steps
+				continue
+			for G in xrange(input_grid_steps):
+				if not G:
+					i += input_grid_steps
+					continue
+				for B in xrange(input_grid_steps):
+					if not B:
+						i += 1
+						continue
+					index = i * samples_per_pixel * bytes_per_sample
+					BGR = self.LUTDATA[index:index + bytes_per_pixel]
+					RGB = BGR[::-1]  # BGR little-endian to RGB big-endian byte order
+					io.write(RGB)
+					i += 1
+		io.write(A2B0.tagData[-output_bytes:])  # Append output curves
+		io.seek(0)
+		link.tags.A2B0 = ICCP.ICCProfileTag(io.read(), "A2B0")
+
+		link.write(stream)
+
+		if isinstance(stream_or_filename, basestring):
+			stream.close()
+
+	def write_tiff(self, stream_or_filename=None):
+		"""
+		Write 3D LUT to TIFF file.
+		
+		"""
+		stream = self._get_stream(stream_or_filename, ".tif")
+
+		# Write image data
+		# XXX Currently only 8 or 16 bit RGB data is supported
+		samples_per_pixel = 3  # RGB
+		bytes_per_sample = self.outputBitDepth / 8
+		bytes_per_pixel = samples_per_pixel * bytes_per_sample
+		w = 2 ** self.inputBitDepth[0]  # Assume equal bitdepth for R, G, B
+		h = w * w
+		stream.write(tiff_get_header(w, h, samples_per_pixel,
+									 self.outputBitDepth))
+		entries = self.lutUncompressedSize / samples_per_pixel / bytes_per_sample
+		for i in xrange(entries):
+			index = i * samples_per_pixel * bytes_per_sample
+			BGR = self.LUTDATA[index:index + bytes_per_pixel]
+			RGB = BGR[::-1]  # BGR little-endian to RGB big-endian byte order
+			stream.write(RGB)
+
 		if isinstance(stream_or_filename, basestring):
 			stream.close()
 
 
-class MadTPG(object):
+class MadTPGBase(object):
+
+	""" Generic pattern generator compatibility layer """
+
+	def wait(self):
+		self.connect(method2=CM_StartLocalInstance)
+
+	def disconnect_client(self):
+		self.disconnect()
+
+	def send(self, rgb=(0, 0, 0), bgrgb=(0, 0, 0), bits=None,
+			 use_video_levels=None, x=0, y=0, w=1, h=1):
+		cfg = self.get_pattern_config()
+		if cfg:
+			self.set_pattern_config(int(round((w + h) / 2.0 * 100)),
+									int(round(sum(bgrgb) / 3.0 * 100)), cfg[2],
+									cfg[3])
+		self.show_rgb(*rgb + bgrgb)
+
+
+class MadTPG(MadTPGBase):
 
 	""" Minimal madTPG controller class """
 
 	def __init__(self):
+		MadTPGBase.__init__(self)
 		self._connection_callbacks = []
 
 		# We only expose stuff we might actually use.
+		# Also, as the HDR 3D LUT install API of madVR is relatively recent
+		# (September 2017), we do not require it.
 
 		# Find madHcNet32.dll
 		clsid = "{E1A8B82A-32CE-4B0D-BE0D-AA68C772E423}"
@@ -219,11 +363,19 @@ class MadTPG(object):
 					prefix = "AutoNet"
 				else:
 					prefix = "madVR"
-				getattr(self.mad, prefix + "_" + methodname).restype = ctypes.c_bool
+				method = getattr(self.mad, prefix + "_" + methodname, None)
+				if not method and not methodname.startswith("LoadHdr3dlut"):
+					raise AttributeError(prefix + "_" + methodname)
+				method.restype = ctypes.c_bool
 
 			# Set expected argument types
 			self.mad.madVR_ShowRGB.argtypes = [ctypes.c_double] * 3
 			self.mad.madVR_ShowRGBEx.argtypes = [ctypes.c_double] * 6
+			if hasattr(self.mad, "madVR_LoadHdr3dlutFile"):
+				self.mad.madVR_LoadHdr3dlutFile.argtypes = [ctypes.wintypes.LPWSTR,
+															ctypes.wintypes.BOOL,
+															ctypes.c_int,
+															ctypes.c_bool]
 		except AttributeError:
 			raise RuntimeError(lang.getstr("madhcnet.outdated",
 										   tuple(reversed(os.path.split(self.dllpath))) +
@@ -330,7 +482,7 @@ class MadTPG(object):
 		return self.dllpath
 
 
-class MadTPG_Net(object):
+class MadTPG_Net(MadTPGBase):
 
 	""" Implementation of madVR network protocol in pure python """
 
@@ -338,13 +490,17 @@ class MadTPG_Net(object):
 	# (tcp.dstport != 1900 and tcp.dstport != 443) or (udp.dstport != 1900 and udp.dstport != 137 and udp.dstport != 138 and udp.dstport != 5355 and udp.dstport != 547 and udp.dstport != 10111)
 
 	def __init__(self):
+		MadTPGBase.__init__(self)
 		self._cast_sockets = {}
 		self._casts = []
 		self._client_sockets = OrderedDict()
 		self._commandno = 0
 		self._commands = {}
-		hostname = socket.gethostname()
-		self._host = socket.gethostbyname(hostname)
+		self._host = get_network_addr()
+		try:
+			hostname = get_valid_host()[0]
+		except socket.error:
+			hostname = self._host
 		self._incoming = {}
 		self._ips = [i[4][0] for i in socket.getaddrinfo(hostname, None)]
 		self._pid = 0
@@ -383,6 +539,8 @@ class MadTPG_Net(object):
 				sock.bind(("", port))
 				sock.listen(1)
 				thread = threading.Thread(target=self._conn_accept_handler,
+										  name="madVR.ConnectionHandler[%s]" %
+											   port,
 										  args=(sock, "", port))
 				self._threads.append(thread)
 				thread.start()
@@ -399,6 +557,8 @@ class MadTPG_Net(object):
 			try:
 				sock.bind(("", port))
 				thread = threading.Thread(target=self._cast_receive_handler,
+										  name="madVR.BroadcastHandler[%s:%s]" %
+											   (self.broadcast_ip, port),
 										  args=(sock, self.broadcast_ip, port))
 				self._threads.append(thread)
 				thread.start()
@@ -419,6 +579,8 @@ class MadTPG_Net(object):
 			try:
 				sock.bind(("", port))
 				thread = threading.Thread(target=self._cast_receive_handler,
+										  name="madVR.MulticastHandler[%s:%s]" %
+											   (self.multicast_ip, port),
 										  args=(sock, self.multicast_ip, port))
 				self._threads.append(thread)
 				thread.start()
@@ -489,6 +651,8 @@ class MadTPG_Net(object):
 				else:
 					self._client_sockets[addr] = conn
 					thread = threading.Thread(target=self._receive_handler,
+											  name="madVR.Receiver[%s:%s]" %
+												   addr[:2],
 											  args=(addr, conn, ))
 					self._threads.append(thread)
 					thread.start()
@@ -622,6 +786,8 @@ class MadTPG_Net(object):
 							else:
 								conn = self._get_client_socket(addr[0], c_port)
 								threading.Thread(target=self._connect,
+												 name="madVR.ConnectToInstance[%s:%s]" %
+													  (addr[0], c_port),
 												 args=(conn, addr[0], c_port)).start()
 					else:
 						self._casts.remove(addr)
@@ -738,6 +904,8 @@ class MadTPG_Net(object):
 		for port in self.server_ports:
 			conn = self._get_client_socket(ip, port)
 			threading.Thread(target=self._connect,
+							 name="madVR.ConnectToInstance[%s:%s]" %
+								  (ip, port),
 							 args=(conn, ip, port, timeout / 1000.0)).start()
 		return self._wait_for_client((ip, port), timeout / 1000.0)
 
@@ -768,6 +936,8 @@ class MadTPG_Net(object):
 				safe_print("MadTPG_Net: Connected to %s:%s" % (host, port))
 			sock.settimeout(0)
 			thread = threading.Thread(target=self._receive_handler,
+									  name="madVR.Receiver[%s:%s]" %
+										   (host, port),
 									  args=((host, port), sock, ))
 			self._threads.append(thread)
 			thread.start()
@@ -1147,17 +1317,19 @@ class MadTPG_Net_Sender(object):
 		self.command = command
 
 	def __call__(self, *args, **kwargs):
-		if self.command == "Load3dlutFile":
+		if self.command in ("Load3dlutFile", "LoadHdr3dlutFile"):
 			lut = H3DLUT(args[0])
 			lutdata = lut.LUTDATA
-			self.command = "Load3dlut"
-		elif self.command == "Load3dlutFromArray256":
+			self.command = self.command[:-4]  # Strip 'File' from command name
+		elif self.command in ("Load3dlutFromArray256", "LoadHdr3dlutFromArray256"):
 			lutdata = args[0]
-			self.command = "Load3dlut"
-		if self.command == "Load3dlut":
+			self.command = self.command[:-12]  # Strip 'File' from command name
+		if self.command in ("Load3dlut", "LoadHdr3dlut"):
 			params = struct.pack("<i", args[1])  # Save to settings?
 			params += struct.pack("<i", args[2])  # 3D LUT slot
 			params += lutdata
+			if self.command == "LoadHdr3dlut":
+				params += struct.pack("<i", args[3])  # HDR to SDR?
 		elif self.command == "SetDeviceGammaRamp":
 			params = ""
 			for j in xrange(3):
@@ -1193,7 +1365,7 @@ class MadTPG_Net_Sender(object):
 				bgb = args[5]
 			rgb = r, g, b
 			if not None in (bgr, bgg, bgb):
-				command += "Ex"
+				self.command += "Ex"
 				rgb += (bgr, bgg, bgb)
 			if None in (r, g, b):
 				raise TypeError("show_rgb() takes at least 4 arguments (%i given)" %
