@@ -699,6 +699,7 @@ class SynthICCFrame(BaseFrame):
 											   for v in black],
 											  getcfg("synthprofile.luminance"),
 											  minmll, maxmll,
+											  getcfg("3dlut.hdr_maxmll_alt_clip"),
 											  rolloff=True,
 											  blend_blackpoint=False)
 				else:
@@ -726,21 +727,22 @@ class SynthICCFrame(BaseFrame):
 									  self.worker.lastmsg])
 					quality = getcfg("profile.quality")
 					clutres = {"m": 17, "l": 9}.get(quality, 33)
-					profile.tags.A2B0 = ICCP.create_synthetic_hdr_clut_profile(
+					hdr_clut_profile = ICCP.create_synthetic_hdr_clut_profile(
 						hdr_format,
 						rgb_space, "",
 						getcfg("synthprofile.black_luminance") * (1 - outoffset),
 						getcfg("synthprofile.luminance"), minmll, maxmll,
+						getcfg("3dlut.hdr_maxmll_alt_clip"),
 						1.2, getcfg("3dlut.hdr_ambient_luminance"),
-						clutres=clutres, worker=self.worker,
-						logfile=logfiles).tags.A2B0
+						clutres=clutres, sat=getcfg("3dlut.hdr_sat"),
+						generate_B2A=trc == -2, worker=self.worker,
+						logfile=logfiles)
+					profile.tags.A2B0 = hdr_clut_profile.tags.A2B0
+					if trc == -2:
+						# HLG
+						profile.tags.B2A0 = hdr_clut_profile.tags.B2A0
 				if black != [0, 0, 0] and outoffset and not bpc:
 					profile.apply_black_offset(black)
-				if rolloff or trc == -2:
-					profile.write(path)
-					self.worker.update_profile_B2A(profile, not bpc,
-												   smooth=False,
-												   rgb_space=rgb_space)
 			else:
 				# Grayscale profile
 				if trc == -2084:
@@ -748,12 +750,14 @@ class SynthICCFrame(BaseFrame):
 														(1 - outoffset),
 														getcfg("synthprofile.luminance"),
 														minmll, maxmll,
+														getcfg("3dlut.hdr_maxmll_alt_clip"),
 														rolloff=True)
 				else:
 					# HLG
-					profile.tags.kTRC.set_smpte2084_trc(0,
-														1.2,
-														getcfg("3dlut.hdr_ambient_luminance"))
+					profile.tags.kTRC.set_hlg_trc(0,
+												  getcfg("synthprofile.luminance"),
+												  1.2,
+												  getcfg("3dlut.hdr_ambient_luminance"))
 				if black != [0, 0, 0] and outoffset and not bpc:
 					profile.tags.kTRC.apply_bpc(black[1])
 		elif black != [0, 0, 0]:
@@ -909,6 +913,9 @@ class SynthICCFrame(BaseFrame):
 		self.black_output_offset_intctrl.SetValue(outoffset)
 		self.lut3d_hdr_minmll_ctrl.SetValue(getcfg("3dlut.hdr_minmll"))
 		self.lut3d_hdr_maxmll_ctrl.SetValue(getcfg("3dlut.hdr_maxmll"))
+		self.lut3d_hdr_maxmll_alt_clip_cb.SetValue(not bool(getcfg("3dlut.hdr_maxmll_alt_clip")))
+		self.lut3d_hdr_sat_ctrl.SetValue(int(round(getcfg("3dlut.hdr_sat") * 100)))
+		self.lut3d_hdr_update_sat_val()
 		setcfg("3dlut.hdr_peak_luminance", getcfg("synthprofile.luminance"))
 		self.lut3d_hdr_update_diffuse_white()
 		self.lut3d_hdr_ambient_luminance_ctrl.SetValue(getcfg("3dlut.hdr_ambient_luminance"))
@@ -919,9 +926,12 @@ class SynthICCFrame(BaseFrame):
 		self.lut3d_hdr_maxmll_label.Show(i == 8)  # SMPTE 2084 (PQ)
 		self.lut3d_hdr_maxmll_ctrl.Show(i == 8)  # SMPTE 2084 (PQ)
 		self.lut3d_hdr_maxmll_ctrl_label.Show(i == 8)  # SMPTE 2084 (PQ)
+		self.lut3d_show_hdr_maxmll_alt_clip_ctrl()
 		self.lut3d_hdr_diffuse_white_label.Show(i == 8)  # SMPTE 2084 (PQ)
 		self.lut3d_hdr_diffuse_white_txt.Show(i == 8)  # SMPTE 2084 (PQ)
 		self.lut3d_hdr_diffuse_white_txt_label.Show(i == 8)  # SMPTE 2084 (PQ)
+		sizer = self.lut3d_hdr_sat_ctrl.ContainingSizer
+		sizer.ShowItems(i in (7, 8))  # SMPTE 2084 (PQ)
 		self.lut3d_hdr_ambient_luminance_label.Show(i == 2)  # HLG
 		self.lut3d_hdr_ambient_luminance_ctrl.Show(i == 2)  # HLG
 		self.lut3d_hdr_ambient_luminance_ctrl_label.Show(i == 2)  # HLG
